/***********************************************************************/
/*                                                                     */
/*   svm_loqo.c                                                        */
/*                                                                     */
/*   Interface to the PR_LOQO optimization package for SVM.            */
/*                                                                     */
/*   Author: Thorsten Joachims                                         */
/*   Date: 17.08.98                                                    */
/*                                                                     */
/*   Copyright (c) 1998  Universitaet Dortmund - All rights reserved   */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/

# include <math.h>
# include "pr_loqo/pr_loqo.h"

/* Common Block Declarations */

long verbosity;

/* /////////////////////////////////////////////////////////////// */

int nx = 400; /* Maximum size of QP-subproblems */

/* /////////////////////////////////////////////////////////////// */

double *optimize_qp();
double svm_c;
double *opt_ce,*opt_ce0,*opt_g,*opt_g0,*opt_xinit;
long   opt_n;
double opt_precision;
double *primal=0,*dual=0,*opt_low,*opt_up;
double epsilon_crit;
double init_margin=0.15;
long   init_iter=500,precision_violations=0;

/* /////////////////////////////////////////////////////////////// */

void *my_malloc();
long max();

double *optimize_qp()
/* start the optimizer and return the optimal values */
{
  register long i,j,result;
  double margin,obj_before,obj_after;
  double sigdig,dist,epsilon_loqo;
  int iter;

  if(!primal) { /* allocate memory at first call */
    primal=(double *)my_malloc(sizeof(double)*nx*3);
    dual=(double *)my_malloc(sizeof(double)*(nx*2+1));
    opt_low=(double *)my_malloc(sizeof(double)*nx);
    opt_up=(double *)my_malloc(sizeof(double)*nx);
    for(i=0;i<nx;i++) {
      opt_low[i]=0;
      opt_up[i]=svm_c;
    }
  }

  if(verbosity>=4) { /* really verbose */
    printf("\n\n");
    for(i=0;i<opt_n;i++) {
      printf("%f: ",opt_g0[i]);
      for(j=0;j<opt_n;j++) {
	printf("%f ",opt_g[i*opt_n+j]);
      }
      printf(": a=%.30f",opt_xinit[i]);
      printf(": y=%f\n",opt_ce[i]);
    }
    printf("\n");
  }

  obj_before=0; /* calculate objective before optimization */
  for(i=0;i<opt_n;i++) {
    obj_before+=(opt_g0[i]*opt_xinit[i]);
    obj_before+=(0.5*opt_xinit[i]*opt_xinit[i]*opt_g[i*opt_n+i]);
    for(j=0;j<i;j++) {
      obj_before+=(opt_xinit[j]*opt_xinit[i]*opt_g[j*opt_n+i]);
    }
  }

  result=STILL_RUNNING;
  opt_ce0[0]*=(-1.0);
  /* Run pr_loqo. If a run fails, try again with parameters which lead */
  /* to a slower, but more robust setting. */
  for(margin=init_margin,iter=init_iter;
      (margin<=0.9999999) && (result!=OPTIMAL_SOLUTION);) {
    sigdig=-log10(opt_precision);
    result=pr_loqo((int)opt_n,(int)1,(double *)opt_g0,(double *)opt_g,
		   (double *)opt_ce,(double *)opt_ce0,
		   (double *)opt_low,(double *)opt_up,
		   (double *)primal,(double *)dual, 
		   /* (int)max((long)(result!=STILL_RUNNING),verbosity-2), */
		   (int)(verbosity-2),
		   (double)sigdig,(int)iter, 
		   (double)margin,(double)svm_c/4.0,(int)0);
    
    if(isnan(dual[0])) {     /* check for choldc problem */
      if(verbosity>=2) {
	printf("NOTICE: Restarting PR_LOQO with more conservative parameters.\n");
      }
      if(init_margin<0.80) { /* become more conservative in general */
	init_margin=(4.0*margin+1.0)/5.0;
      }
      margin=(margin+1.0)/2.0;
      opt_precision*=10.0;   /* reduce precision */
      if(verbosity>=2) {
	printf("NOTICE: Reducing precision of PR_LOQO.\n");
      }
    }
    else if(result!=OPTIMAL_SOLUTION) {
      iter+=2000; 
      init_iter+=10;
      opt_precision*=10.0;   /* reduce precision */
      if(verbosity>=2) {
	printf("NOTICE: Reducing precision of PR_LOQO.\n");
      }      
    }
  }
  
  /* Check the precision of the alphas. If results of current optimization */
  /* violate KT-Conditions, relax the epsilon on the bounds on alphas. */
  epsilon_loqo=1E-10;
  for(i=0;i<opt_n;i++) {
    dist=-dual[0]*opt_ce[i]; /* Thanks to Alex Smola for this hint */
    dist+=(opt_g0[i]+1.0);
    for(j=0;j<i;j++) {
      dist+=(primal[j]*opt_g[j*opt_n+i]);
    }
    for(j=i;j<opt_n;j++) {
      dist+=(primal[j]*opt_g[i*opt_n+j]);
    }
    /* printf("LOQO: a[%d]=%f, dist=%f, b=%f\n",i,primal[i],dist,dual[0]); */
    if((primal[i]<(svm_c-epsilon_loqo)) && (dist < (1.0-epsilon_crit))) {
      epsilon_loqo=(svm_c-primal[i])*2.0;
    }
    else if((primal[i]>(0+epsilon_loqo)) && (dist > (1.0+epsilon_crit))) {
      epsilon_loqo=primal[i]*2.0;
    }
  }

  for(i=0;i<opt_n;i++) {  /* clip alphas to bounds */
    if(primal[i]<=(0+epsilon_loqo)) {
      primal[i]=0;
    }
    else if(primal[i]>=(svm_c-epsilon_loqo)) {
      primal[i]=svm_c;
    }
  }

  obj_after=0;  /* calculate objective after optimization */
  for(i=0;i<opt_n;i++) {
    obj_after+=(opt_g0[i]*primal[i]);
    obj_after+=(0.5*primal[i]*primal[i]*opt_g[i*opt_n+i]);
    for(j=0;j<i;j++) {
      obj_after+=(primal[j]*primal[i]*opt_g[j*opt_n+i]);
    }
  }

  if(obj_after >= obj_before) { /* check whether there was progress */
    opt_precision/=100.0;
    precision_violations++;
    if(verbosity>=2) {
      printf("NOTICE: Increasing Precision of PR_LOQO.\n");
    }
  }

  if(precision_violations > 50) {
    epsilon_crit*=10.0;
    if(verbosity>=1) {
      printf("\nWARNING: Relaxing epsilon on KT-Conditions.\n");
    }
  }	  

  if(result!=OPTIMAL_SOLUTION) {
    printf("\nERROR: PR_LOQO did not converge. \n");
    return(opt_xinit);
  }
  else {
    return(primal);
  }
}

